{*******************************************************}
{                                                       }
{         Vladimir Gaitanoff Delphi VCL Library         }
{         MySQL registration unit                       }
{                                                       }
{         Copyright (c) 1997, 1999                      }
{                                                       }
{*******************************************************}

{$I VG.INC}
{$D-,L-}

unit vgMySQLReg;

interface

const
  SMySQLPage = 'VG MySQL';

procedure Register;

implementation
uses Classes, vgUtils, vgMySQL, DsgnIntf;

{$R vgMySQL.dcr}

type
{ TMySQLNamesProperty }
  TMySQLNamesProperty = class(TStringProperty)
  public
    function GetAttributes: TPropertyAttributes; override;
    procedure GetValueList(List: TStrings); virtual;
    procedure GetValues(Proc: TGetStrProc); override;
  end;

{ TMySQLDatabaseNameProperty }
  TMySQLDatabaseNameProperty = class(TMySQLNamesProperty)
  public
    procedure GetValueList(List: TStrings); override;
  end;

{ TMySQLTableNameProperty }
  TMySQLTableNameProperty = class(TMySQLNamesProperty)
  public
    procedure GetValueList(List: TStrings); override;
  end;

{ TMySQLNamesProperty }
function TMySQLNamesProperty.GetAttributes: TPropertyAttributes;
begin
  Result := [paValueList, paSortList, paMultiSelect];
end;

procedure TMySQLNamesProperty.GetValueList(List: TStrings);
begin
  { nothing }
end;

procedure TMySQLNamesProperty.GetValues(Proc: TGetStrProc);
var
  List: TStrings;
begin
  List := TStringList.Create;
  try
    GetValueList(List);
    EnumStrings(List, Proc);
  finally
    List.Free;
  end;
end;

{ TMySQLDatabaseNameProperty }
procedure TMySQLDatabaseNameProperty.GetValueList(List: TStrings);
var
  Connection: TMySQLConnection;
begin
  Connection := GetComponent(0) as TMySQLConnection;
  Connection.GetObjectList(otDatabase, '', List);
end;

{ TMySQLTableNameProperty }
procedure TMySQLTableNameProperty.GetValueList(List: TStrings);
var
  Table: TMySQLTable;
begin
  Table := GetComponent(0) as TMySQLTable;
  if Assigned(Table.Connection) then
    Table.Connection.GetObjectList(otTable, '', List);
end;

procedure Register;
begin
  RegisterComponents(SMySQLPage, [TMySQLConnection, TMySQLDataSet, TMySQLTable, TMySQLQuery, TMySQLObjects]);
  RegisterPropertyEditor(TypeInfo(string), TMySQLConnection, 'DatabaseName', TMySQLDatabaseNameProperty);
  RegisterPropertyEditor(TypeInfo(string), TMySQLTable, 'TableName', TMySQLTableNameProperty);
end;

end.
